@@ -50,7 +50,10 @@
 #include "lv2/lv2plug.in/ns/ext/event/event.h"
 #include "lv2/lv2plug.in/ns/ext/options/options.h"
 #include "lv2/lv2plug.in/ns/ext/parameters/parameters.h"
+#include "lv2/lv2plug.in/ns/ext/patch/patch.h"
+#include "lv2/lv2plug.in/ns/ext/port-groups/port-groups.h"
 #include "lv2/lv2plug.in/ns/ext/presets/presets.h"
+#include "lv2/lv2plug.in/ns/ext/state/state.h"
 #include "lv2/lv2plug.in/ns/ext/time/time.h"
 #include "lv2/lv2plug.in/ns/ext/uri-map/uri-map.h"
 #include "lv2/lv2plug.in/ns/ext/urid/urid.h"
@@ -82,20 +85,35 @@
 #    define REALTIME
 #endif
 
+/* Size factor for UI ring buffers.  The ring size is a few times the size of
+   an event output to give the UI a chance to keep up.  Experiments with Ingen,
+   which can highly saturate its event output, led me to this value.  It
+   really ought to be enough for anybody(TM).
+*/
+#define N_BUFFER_CYCLES 16
+
 ZixSem exit_sem;  /**< Exit semaphore */
 
 static LV2_URID
 map_uri(LV2_URID_Map_Handle handle,
         const char*         uri)
 {
-	return symap_map(((Jalv*)handle)->symap, uri);
+	Jalv* jalv = (Jalv*)handle;
+	zix_sem_wait(&jalv->symap_lock);
+	const LV2_URID id = symap_map(jalv->symap, uri);
+	zix_sem_post(&jalv->symap_lock);
+	return id;
 }
 
 static const char*
 unmap_uri(LV2_URID_Unmap_Handle handle,
           LV2_URID              urid)
 {
-	return symap_unmap(((Jalv*)handle)->symap, urid);
+	Jalv* jalv = (Jalv*)handle;
+	zix_sem_wait(&jalv->symap_lock);
+	const char* uri = symap_unmap(jalv->symap, urid);
+	zix_sem_post(&jalv->symap_lock);
+	return uri;
 }
 
 /**
@@ -106,7 +124,11 @@ uri_to_id(LV2_URI_Map_Callback_Data callback_data,
           const char*               map,
           const char*               uri)
 {
-	return symap_map(((Jalv*)callback_data)->symap, uri);
+	Jalv* jalv = (Jalv*)callback_data;
+	zix_sem_wait(&jalv->symap_lock);
+	const LV2_URID id = symap_map(jalv->symap, uri);
+	zix_sem_post(&jalv->symap_lock);
+	return id;
 }
 
 #define NS_EXT "http://lv2plug.in/ns/ext/"
@@ -120,6 +142,7 @@ static LV2_Feature make_path_feature = { LV2_STATE__makePath, NULL };
 static LV2_Feature schedule_feature  = { LV2_WORKER__schedule, NULL };
 static LV2_Feature log_feature       = { LV2_LOG__log, NULL };
 static LV2_Feature options_feature   = { LV2_OPTIONS__options, NULL };
+static LV2_Feature def_state_feature = { LV2_STATE__loadDefaultState, NULL };
 
 /** These features have no data */
 static LV2_Feature buf_size_features[3] = {
@@ -127,12 +150,13 @@ static LV2_Feature buf_size_features[3] = {
 	{ LV2_BUF_SIZE__fixedBlockLength, NULL },
 	{ LV2_BUF_SIZE__boundedBlockLength, NULL } };
 
-const LV2_Feature* features[12] = {
+const LV2_Feature* features[13] = {
 	&uri_map_feature, &map_feature, &unmap_feature,
 	&make_path_feature,
 	&schedule_feature,
 	&log_feature,
 	&options_feature,
+	&def_state_feature,
 	&buf_size_features[0],
 	&buf_size_features[1],
 	&buf_size_features[2],
@@ -162,7 +186,7 @@ die(const char* msg)
 /**
    Create a port structure from data description.  This is called before plugin
    and Jack instantiation.  The remaining instance-specific setup
-   (e.g. buffers) is done later in expose_port().
+   (e.g. buffers) is done later in activate_port().
 */
 static void
 create_port(Jalv*    jalv,
@@ -174,6 +198,7 @@ create_port(Jalv*    jalv,
 	port->lilv_port = lilv_plugin_get_port_by_index(jalv->plugin, port_index);
 	port->jack_port = NULL;
 	port->evbuf     = NULL;
+	port->buf_size  = 0;
 	port->index     = port_index;
 	port->control   = 0.0f;
 	port->flow      = FLOW_UNKNOWN;
@@ -182,9 +207,8 @@ create_port(Jalv*    jalv,
 	const LilvNode* symbol = lilv_port_get_symbol(jalv->plugin,
 	                                              port->lilv_port);
 
-	const bool optional = lilv_port_has_property(jalv->plugin,
-	                                             port->lilv_port,
-	                                             jalv->nodes.lv2_connectionOptional);
+	const bool optional = lilv_port_has_property(
+		jalv->plugin, port->lilv_port, jalv->nodes.lv2_connectionOptional);
 
 	/* Set the port flow (input or output) */
 	if (lilv_port_is_a(jalv->plugin, port->lilv_port, jalv->nodes.lv2_InputPort)) {
@@ -215,6 +239,15 @@ create_port(Jalv*    jalv,
 		die("Mandatory port has unknown data type");
 	}
 
+	LilvNode* min_size = lilv_port_get(
+		jalv->plugin, port->lilv_port, jalv->nodes.rsz_minimumSize);
+	if (min_size && lilv_node_is_int(min_size)) {
+		port->buf_size = lilv_node_as_int(min_size);
+		jalv->opts.buffer_size = MAX(
+			jalv->opts.buffer_size, port->buf_size * N_BUFFER_CYCLES);
+	}
+	lilv_node_free(min_size);
+
 	const size_t sym_len = strlen(lilv_node_as_string(symbol));
 	if (sym_len > jalv->longest_sym) {
 		jalv->longest_sym = sym_len;
@@ -237,6 +270,12 @@ jalv_create_ports(Jalv* jalv)
 		create_port(jalv, i, default_values[i]);
 	}
 
+	const LilvPort* control_input = lilv_plugin_get_port_by_designation(
+		jalv->plugin, jalv->nodes.lv2_InputPort, jalv->nodes.lv2_control);
+	if (control_input) {
+		jalv->control_in = lilv_port_get_index(jalv->plugin, control_input);
+	}
+
 	free(default_values);
 }
 
@@ -251,8 +290,11 @@ jalv_allocate_port_buffers(Jalv* jalv)
 		switch (port->type) {
 		case TYPE_EVENT:
 			lv2_evbuf_free(port->evbuf);
+			const size_t buf_size = (port->buf_size > 0)
+				? port->buf_size
+				: jalv->midi_buf_size;
 			port->evbuf = lv2_evbuf_new(
-				jalv->midi_buf_size,
+				buf_size,
 				port->old_api ? LV2_EVBUF_EVENT : LV2_EVBUF_ATOM,
 				jalv->map.map(jalv->map.handle,
 				              lilv_node_as_string(jalv->nodes.atom_Chunk)),
@@ -362,7 +404,8 @@ jack_process_cb(jack_nframes_t nframes, void* data)
 
 	/* If transport state is not as expected, then something has changed */
 	const bool xport_changed = (rolling != jalv->rolling ||
-	                            pos.frame != jalv->position);
+	                            pos.frame != jalv->position ||
+	                            pos.beats_per_minute != jalv->bpm);
 
 	uint8_t   pos_buf[256];
 	LV2_Atom* lv2_pos = (LV2_Atom*)pos_buf;
@@ -381,9 +424,9 @@ jack_process_cb(jack_nframes_t nframes, void* data)
 			lv2_atom_forge_float(
 				forge, pos.beat - 1 + (pos.tick / pos.ticks_per_beat));
 			lv2_atom_forge_property_head(forge, jalv->urids.time_bar, 0);
-			lv2_atom_forge_float(forge, pos.bar - 1);
+			lv2_atom_forge_long(forge, pos.bar - 1);
 			lv2_atom_forge_property_head(forge, jalv->urids.time_beatUnit, 0);
-			lv2_atom_forge_float(forge, pos.beat_type);
+			lv2_atom_forge_int(forge, pos.beat_type);
 			lv2_atom_forge_property_head(forge, jalv->urids.time_beatsPerBar, 0);
 			lv2_atom_forge_float(forge, pos.beats_per_bar);
 			lv2_atom_forge_property_head(forge, jalv->urids.time_beatsPerMinute, 0);
@@ -401,6 +444,7 @@ jack_process_cb(jack_nframes_t nframes, void* data)
 
 	/* Update transport state to expected values for next cycle */
 	jalv->position = rolling ? pos.frame + nframes : pos.frame;
+	jalv->bpm      = pos.beats_per_minute;
 	jalv->rolling  = rolling;
 
 	switch (jalv->play_state) {
@@ -675,7 +719,7 @@ jalv_ui_write(SuilController controller,
 
 	if (protocol != 0 && protocol != jalv->urids.atom_eventTransfer) {
 		fprintf(stderr, "UI write with unsupported protocol %d (%s)\n",
-		        protocol, symap_unmap(jalv->symap, protocol));
+		        protocol, unmap_uri(jalv, protocol));
 		return;
 	}
 
@@ -686,12 +730,9 @@ jalv_ui_write(SuilController controller,
 	}
 
 	if (jalv->opts.dump && protocol == jalv->urids.atom_eventTransfer) {
-		SerdNode s = serd_node_from_string(SERD_BLANK, USTR("msg"));
-		SerdNode p = serd_node_from_string(SERD_URI, USTR(NS_RDF "value"));
-
 		const LV2_Atom* atom = (const LV2_Atom*)buffer;
 		char*           str  = sratom_to_turtle(
-			jalv->sratom, &jalv->unmap, "jalv:", &s, &p,
+			jalv->sratom, &jalv->unmap, "jalv:", NULL, NULL,
 			atom->type, atom->size, LV2_ATOM_BODY(atom));
 		printf("\n## UI => Plugin (%u bytes) ##\n%s\n", atom->size, str);
 		free(str);
@@ -711,17 +752,24 @@ jalv_emit_ui_events(Jalv* jalv)
 {
 	ControlChange ev;
 	const size_t  space = jack_ringbuffer_read_space(jalv->plugin_events);
-	for (size_t i = 0; i < space; i += sizeof(ev) + ev.size) {
+	for (size_t i = 0;
+	     i + sizeof(ev) + sizeof(float) <= space;
+	     i += sizeof(ev) + ev.size) {
+		// Read event header to get the size
 		jack_ringbuffer_read(jalv->plugin_events, (char*)&ev, sizeof(ev));
-		char buf[ev.size];
+
+		// Resize read buffer if necessary
+		jalv->ui_event_buf = realloc(jalv->ui_event_buf, ev.size);
+		void* const buf = jalv->ui_event_buf;
+
+		// Read event body
 		jack_ringbuffer_read(jalv->plugin_events, buf, ev.size);
 
 		if (jalv->opts.dump && ev.protocol == jalv->urids.atom_eventTransfer) {
-			SerdNode  s    = serd_node_from_string(SERD_BLANK, USTR("msg"));
-			SerdNode  p    = serd_node_from_string(SERD_URI, USTR(NS_RDF "value"));
+			// Dump event in Turtle to the console
 			LV2_Atom* atom = (LV2_Atom*)buf;
 			char*     str  = sratom_to_turtle(
-				jalv->sratom, &jalv->unmap, "jalv:", &s, &p,
+				jalv->ui_sratom, &jalv->unmap, "jalv:", NULL, NULL,
 				atom->type, atom->size, LV2_ATOM_BODY(atom));
 			printf("\n## Plugin => UI (%u bytes) ##\n%s\n", atom->size, str);
 			free(str);
@@ -753,6 +801,7 @@ main(int argc, char** argv)
 	jalv.block_length  = 4096;  // Should be set by jack_buffer_size_cb
 	jalv.midi_buf_size = 1024;  // Should be set by jack_buffer_size_cb
 	jalv.play_state    = JALV_PAUSED;
+	jalv.bpm           = 120.0f;
 
 	if (jalv_init(&argc, &argv, &jalv.opts)) {
 		return EXIT_FAILURE;
@@ -763,6 +812,7 @@ main(int argc, char** argv)
 	}
 
 	jalv.symap = symap_new();
+	zix_sem_init(&jalv.symap_lock, 1);
 	uri_map.callback_data = &jalv;
 
 	jalv.map.handle  = &jalv;
@@ -775,7 +825,8 @@ main(int argc, char** argv)
 
 	lv2_atom_forge_init(&jalv.forge, &jalv.map);
 
-	jalv.sratom = sratom_new(&jalv.map);
+	jalv.sratom    = sratom_new(&jalv.map);
+	jalv.ui_sratom = sratom_new(&jalv.map);
 
 	jalv.midi_event_id = uri_to_id(
 		&jalv, "http://lv2plug.in/ns/ext/event", LV2_MIDI__MidiEvent);
@@ -789,6 +840,9 @@ main(int argc, char** argv)
 	jalv.urids.log_Trace            = symap_map(jalv.symap, LV2_LOG__Trace);
 	jalv.urids.midi_MidiEvent       = symap_map(jalv.symap, LV2_MIDI__MidiEvent);
 	jalv.urids.param_sampleRate     = symap_map(jalv.symap, LV2_PARAMETERS__sampleRate);
+	jalv.urids.patch_Set            = symap_map(jalv.symap, LV2_PATCH__Set);
+	jalv.urids.patch_property       = symap_map(jalv.symap, LV2_PATCH__property);
+	jalv.urids.patch_value          = symap_map(jalv.symap, LV2_PATCH__value);
 	jalv.urids.time_Position        = symap_map(jalv.symap, LV2_TIME__Position);
 	jalv.urids.time_bar             = symap_map(jalv.symap, LV2_TIME__bar);
 	jalv.urids.time_barBeat         = symap_map(jalv.symap, LV2_TIME__barBeat);
@@ -797,6 +851,7 @@ main(int argc, char** argv)
 	jalv.urids.time_beatsPerMinute  = symap_map(jalv.symap, LV2_TIME__beatsPerMinute);
 	jalv.urids.time_frame           = symap_map(jalv.symap, LV2_TIME__frame);
 	jalv.urids.time_speed           = symap_map(jalv.symap, LV2_TIME__speed);
+	jalv.urids.ui_updateRate        = symap_map(jalv.symap, LV2_UI__updateRate);
 
 #ifdef _WIN32
 	jalv.temp_dir = jalv_strdup("jalvXXXXXX");
@@ -841,9 +896,13 @@ main(int argc, char** argv)
 	jalv.nodes.lv2_InputPort          = lilv_new_uri(world, LV2_CORE__InputPort);
 	jalv.nodes.lv2_OutputPort         = lilv_new_uri(world, LV2_CORE__OutputPort);
 	jalv.nodes.lv2_connectionOptional = lilv_new_uri(world, LV2_CORE__connectionOptional);
+	jalv.nodes.lv2_control            = lilv_new_uri(world, LV2_CORE__control);
+	jalv.nodes.lv2_name               = lilv_new_uri(world, LV2_CORE__name);
 	jalv.nodes.midi_MidiEvent         = lilv_new_uri(world, LV2_MIDI__MidiEvent);
+	jalv.nodes.pg_group               = lilv_new_uri(world, LV2_PORT_GROUPS__group);
 	jalv.nodes.pset_Preset            = lilv_new_uri(world, LV2_PRESETS__Preset);
 	jalv.nodes.rdfs_label             = lilv_new_uri(world, LILV_NS_RDFS "label");
+	jalv.nodes.rsz_minimumSize        = lilv_new_uri(world, LV2_RESIZE_PORT__minimumSize);
 	jalv.nodes.work_interface         = lilv_new_uri(world, LV2_WORKER__interface);
 	jalv.nodes.work_schedule          = lilv_new_uri(world, LV2_WORKER__schedule);
 	jalv.nodes.end                    = NULL;
@@ -896,6 +955,13 @@ main(int argc, char** argv)
 			return EXIT_FAILURE;
 		}
 	}
+	lilv_nodes_free(req_feats);
+
+	if (!state) {
+		/* Not restoring state, load the plugin as a preset to get default */
+		state = lilv_state_new_from_world(
+			jalv.world, &jalv.map, lilv_plugin_get_uri(jalv.plugin));
+	}
 
 	/* Get a plugin UI */
 	const char* native_ui_type_uri = jalv_native_ui_type(&jalv);
@@ -977,9 +1043,25 @@ main(int argc, char** argv)
 		   should be able to keep up with 4 cycles, and tests show this works
 		   for me, but this value might need increasing to avoid overflows.
 		*/
-		jalv.opts.buffer_size = jalv.midi_buf_size * 4;
+		jalv.opts.buffer_size = jalv.midi_buf_size * N_BUFFER_CYCLES;
 	}
 
+	if (jalv.opts.update_rate == 0.0) {
+		/* Calculate a reasonable UI update frequency. */
+		jalv.ui_update_hz = (float)jalv.sample_rate / jalv.midi_buf_size * 2.0f;
+		jalv.ui_update_hz = MAX(25.0f, jalv.ui_update_hz);
+	} else {
+		/* Use user-specified UI update rate. */
+		jalv.ui_update_hz = jalv.opts.update_rate;
+		jalv.ui_update_hz = MAX(1.0f, jalv.ui_update_hz);
+	}
+
+	/* The UI can only go so fast, clamp to reasonable limits */
+	jalv.ui_update_hz     = MIN(60, jalv.ui_update_hz);
+	jalv.opts.buffer_size = MAX(4096, jalv.opts.buffer_size);
+	fprintf(stderr, "Comm buffers: %d bytes\n", jalv.opts.buffer_size);
+	fprintf(stderr, "Update rate:  %.01f Hz\n", jalv.ui_update_hz);
+
 	/* Build options array to pass to plugin */
 	const LV2_Options_Option options[] = {
 		{ LV2_OPTIONS_INSTANCE, 0, jalv.urids.param_sampleRate,
@@ -990,20 +1072,13 @@ main(int argc, char** argv)
 		  sizeof(int32_t), jalv.urids.atom_Int, &jalv.block_length },
 		{ LV2_OPTIONS_INSTANCE, 0, jalv.urids.bufsz_sequenceSize,
 		  sizeof(int32_t), jalv.urids.atom_Int, &jalv.midi_buf_size },
+		{ LV2_OPTIONS_INSTANCE, 0, jalv.urids.ui_updateRate,
+		  sizeof(float), jalv.urids.atom_Float, &jalv.ui_update_hz },
 		{ LV2_OPTIONS_INSTANCE, 0, 0, 0, 0, NULL }
 	};
 	
 	options_feature.data = &options;
 
-	/* Calculate theoretical UI update frequency. */
-	jalv.ui_update_hz = (double)jalv.sample_rate / jalv.midi_buf_size * 2.0;
-
-	/* The UI can only go so fast, clamp to reasonable limits */
-	jalv.ui_update_hz     = MIN(60, jalv.ui_update_hz);
-	jalv.opts.buffer_size = MAX(4096, jalv.opts.buffer_size);
-	fprintf(stderr, "Comm buffers: %d bytes\n", jalv.opts.buffer_size);
-	fprintf(stderr, "Update rate:  %d Hz\n", jalv.ui_update_hz);
-
 	/* Create Plugin <=> UI communication buffers */
 	jalv.ui_events     = jack_ringbuffer_create(jalv.opts.buffer_size);
 	jalv.plugin_events = jack_ringbuffer_create(jalv.opts.buffer_size);
@@ -1093,8 +1168,10 @@ main(int argc, char** argv)
 		lilv_node_free(*n);
 	}
 	symap_free(jalv.symap);
+	zix_sem_destroy(&jalv.symap_lock);
 	suil_host_free(jalv.ui_host);
 	sratom_free(jalv.sratom);
+	sratom_free(jalv.ui_sratom);
 	lilv_uis_free(jalv.uis);
 	lilv_world_free(world);
 
@@ -1102,6 +1179,7 @@ main(int argc, char** argv)
 
 	remove(jalv.temp_dir);
 	free(jalv.temp_dir);
+	free(jalv.ui_event_buf);
 
 	return 0;
 }
