@@ -31,6 +31,7 @@
 #include "lv2/lv2plug.in/ns/ext/atom/forge.h"
 #include "lv2/lv2plug.in/ns/ext/log/log.h"
 #include "lv2/lv2plug.in/ns/ext/midi/midi.h"
+#include "lv2/lv2plug.in/ns/ext/resize-port/resize-port.h"
 #include "lv2/lv2plug.in/ns/ext/state/state.h"
 #include "lv2/lv2plug.in/ns/ext/urid/urid.h"
 #include "lv2/lv2plug.in/ns/ext/worker/worker.h"
@@ -67,6 +68,7 @@ struct Port {
 	jack_port_t*    jack_port;  ///< For audio/MIDI ports, otherwise NULL
 	LV2_Evbuf*      evbuf;      ///< For MIDI ports, otherwise NULL
 	void*           widget;     ///< Control widget, if applicable
+	size_t          buf_size;   ///< Custom buffer size, or 0
 	uint32_t        index;      ///< Port index
 	float           control;    ///< For control ports, otherwise 0.0f
 	bool            old_api;    ///< True for event, false for atom
@@ -86,6 +88,7 @@ typedef struct {
 	char*    uuid;
 	char*    load;
 	uint32_t buffer_size;
+	double   update_rate;
 	bool     dump;
 	bool     generic_ui;
 } JalvOptions;
@@ -100,6 +103,9 @@ typedef struct {
 	LV2_URID log_Trace;
 	LV2_URID midi_MidiEvent;
 	LV2_URID param_sampleRate;
+	LV2_URID patch_Set;
+	LV2_URID patch_property;
+	LV2_URID patch_value;
 	LV2_URID time_Position;
 	LV2_URID time_bar;
 	LV2_URID time_barBeat;
@@ -108,6 +114,7 @@ typedef struct {
 	LV2_URID time_beatsPerMinute;
 	LV2_URID time_frame;
 	LV2_URID time_speed;
+	LV2_URID ui_updateRate;
 } JalvURIDs;
 
 typedef struct {
@@ -120,9 +127,13 @@ typedef struct {
 	LilvNode* lv2_InputPort;
 	LilvNode* lv2_OutputPort;
 	LilvNode* lv2_connectionOptional;
+	LilvNode* lv2_control;
+	LilvNode* lv2_name;
 	LilvNode* midi_MidiEvent;
+	LilvNode* pg_group;
 	LilvNode* pset_Preset;
 	LilvNode* rdfs_label;
+	LilvNode* rsz_minimumSize;
 	LilvNode* work_interface;
 	LilvNode* work_schedule;
 	LilvNode* end;  ///< NULL terminator for easy freeing of entire structure
@@ -153,10 +164,13 @@ typedef struct {
 	LV2_URID_Map       map;            ///< URI => Int map
 	LV2_URID_Unmap     unmap;          ///< Int => URI map
 	Sratom*            sratom;         ///< Atom serialiser
-	Symap*             symap;          ///< Symbol (URI) map
+	Sratom*            ui_sratom;      ///< Atom serialiser for UI thread
+	Symap*             symap;          ///< URI map
+	ZixSem             symap_lock;     ///< Lock for URI map
 	jack_client_t*     jack_client;    ///< Jack client
 	jack_ringbuffer_t* ui_events;      ///< Port events from UI
 	jack_ringbuffer_t* plugin_events;  ///< Port events from plugin
+	void*              ui_event_buf;   ///< Buffer for reading UI port events
 	JalvWorker         worker;         ///< Worker thread implementation
 	ZixSem*            done;           ///< Exit semaphore
 	ZixSem             paused;         ///< Paused signal from process thread
@@ -174,13 +188,15 @@ typedef struct {
 	struct Port*       ports;          ///< Port array of size num_ports
 	uint32_t           block_length;   ///< Jack buffer size (block length)
 	size_t             midi_buf_size;  ///< Size of MIDI port buffers
+	uint32_t           control_in;     ///< Index of control input port
 	uint32_t           num_ports;      ///< Size of the two following arrays:
 	uint32_t           longest_sym;    ///< Longest port symbol
-	uint32_t           ui_update_hz;   ///< Frequency of UI updates
+	float              ui_update_hz;   ///< Frequency of UI updates
 	jack_nframes_t     sample_rate;    ///< Sample rate
 	jack_nframes_t     event_delta_t;  ///< Frames since last update sent to UI
 	uint32_t           midi_event_id;  ///< MIDI event class ID in event context
 	jack_nframes_t     position;       ///< Transport position in frames
+	float              bpm;            ///< Transport tempo in beats per minute
 	bool               rolling;        ///< Transport speed (0=stop, 1=play)
 	bool               buf_size_set;   ///< True iff buffer size callback fired
 	bool               exit;           ///< True iff execution is finished
